#include "utils.h"


class CJpegPushStream : public CSourceStream, public CSourceSeeking
{
friend class CJpegPushSource;
private:
	REFERENCE_TIME	m_tpf;
	CMediaType		m_mt;
	LONGLONG		m_i;
	LONGLONG		m_k;
	bool			m_gap;

	CJpegPushStream(HRESULT* phr, CSource* pms)
		: CSourceStream(TEXT("CJpegPushStream"),phr,pms,L"out")
#pragma warning(push)
#pragma warning(disable: 4355) /// 'this' : used in base member initializer list
		, CSourceSeeking(TEXT("CJpegPushStream"),(IPin*)this,phr,pms->pStateLock())
#pragma warning(pop)
		, m_tpf(100 * REFTIME_UNITS_PER_MILLISEC) /// change this to change the frame rate
		, m_i(0i64), m_k(0i64), m_gap(true)
	{
		if(S_OK != *phr) return;

		HRESULT hr = S_OK;

		VIDEOINFOHEADER* vih = (VIDEOINFOHEADER*)m_mt.AllocFormatBuffer(sizeof(*vih));
		if(vih == NULL) { hr = E_OUTOFMEMORY; goto exit; } else ZeroMemory(vih,sizeof(*vih));

		vih->bmiHeader.biSize			= sizeof(vih->bmiHeader);
		vih->bmiHeader.biWidth			= 320; /// change this to change the resolution
		vih->bmiHeader.biHeight			= 240; /// change this to change the resolution
		vih->bmiHeader.biPlanes			= 1;
		vih->bmiHeader.biBitCount		= 24;
		vih->bmiHeader.biCompression	= *(DWORD*)"MJPG";
		vih->bmiHeader.biSizeImage		= 64*1024 + sizeof(RGBQUAD) * vih->bmiHeader.biWidth * vih->bmiHeader.biHeight;
		vih->rcSource.right				= vih->bmiHeader.biWidth;
		vih->rcSource.bottom			= vih->bmiHeader.biHeight;
		vih->rcTarget					= vih->rcSource;
		vih->AvgTimePerFrame			= m_tpf;

		m_mt.SetType(&MEDIATYPE_Video);
		m_mt.SetSubtype(&FOURCCMap(vih->bmiHeader.biCompression));
		m_mt.SetVariableSize();
		m_mt.SetTemporalCompression(FALSE);
		m_mt.SetSampleSize(vih->bmiHeader.biSizeImage);
		m_mt.SetFormatType(&FORMAT_VideoInfo);

	exit:
		*phr = hr;
	}

	/// *** CUnknown *** ///
protected:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(riid == __uuidof(IMediaSeeking)) return this->CSourceSeeking::NonDelegatingQueryInterface(riid,ppv);
		return this->CSourceStream::NonDelegatingQueryInterface(riid,ppv);
	}

	/// *** CSourceStream *** ///
protected:
	HRESULT GetMediaType(CMediaType* pmt)
	{
		return pmt->Set(m_mt);
	}
	HRESULT DecideBufferSize(IMemAllocator* pAlloc, ALLOCATOR_PROPERTIES* lpReq)
	{
		ALLOCATOR_PROPERTIES apReq = { 1L, (long)m_mt.GetSampleSize(), 1L, 0L }, apRes = { 0L, 0L, 0L, 0L };

		HRESULT hr = pAlloc->SetProperties(&apReq,&apRes); if(S_OK != hr) goto exit;

		if(apRes.cBuffers < 1L) { hr = VFW_E_BUFFER_NOTSET; goto exit; }
		if(apRes.cbBuffer < (long)m_mt.GetSampleSize()) { hr = VFW_E_BUFFER_OVERFLOW; goto exit; }
		if(apRes.cbAlign < 1L) { hr = VFW_E_BADALIGN; goto exit; }
		if(apRes.cbPrefix != 0L) { hr = VFW_E_BADALIGN; goto exit; }

	exit:
		return hr;
	}
	HRESULT FillBuffer(IMediaSample* pms)
	{
		HRESULT	hr	= S_OK;
		HANDLE	hf	= INVALID_HANDLE_VALUE;
		TCHAR	fn[MAX_PATH];
		LPBYTE	pb;
		long	cb;

		if(S_OK != (hr = pms->GetPointer(&pb))) goto exit; else cb = pms->GetSize();

		/// change this block to fetch the JPEG data in some other way
		{
			/// change this to change the naming scheme of the JPEG files
			(void)wnsprintf(fn,sizeof(fn)/sizeof(*fn),TEXT(".\\source\\%03lu.jpg"),m_i);

			if(INVALID_HANDLE_VALUE == (hf = CreateFile(fn,FILE_READ_DATA,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL))) {
				DWORD rc = GetLastError();
				hr = ERROR_FILE_NOT_FOUND == rc ? S_FALSE : HRESULT_FROM_WIN32(rc); goto exit;
			}
			if(!ReadFile(hf,pb,(DWORD)cb,(DWORD*)&cb,NULL)) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; }
			if(cb < 1) { hr = HRESULT_FROM_WIN32(ERROR_HANDLE_EOF); goto exit; }
		}

		if(S_OK != (hr = pms->SetActualDataLength(cb))) goto exit;
		if(S_OK != (hr = pms->SetSyncPoint(TRUE))) goto exit;
		if(m_gap) { if(S_OK != (hr = pms->SetDiscontinuity(TRUE))) goto exit; m_gap = false; } /// since we have only keyframes, this is not strictly necessary

		/// timestamps must restart from 0 after a seek, so substract the first frame (m_k) from the current frame (m_i)
		LONGLONG mt1 = m_i - m_k, mt2 = mt1 + 1;
		REFERENCE_TIME rt1 = mt1 * m_tpf, rt2 = rt1 + m_tpf;

		if(S_OK != (hr = pms->SetTime(&rt1,&rt2))) goto exit;
		/// this is not necessary, but easy to provide with a frame-based stream
		if(S_OK != (hr = pms->SetMediaTime(&mt1,&mt2))) goto exit;

		m_i++;

	exit:
		if(hf != INVALID_HANDLE_VALUE) (void)CloseHandle(hf);
		return hr;
	}
	HRESULT OnThreadStartPlay(void)
	{
		m_gap = true;
		return DeliverNewSegment(m_rtStart,m_rtStop,m_dRateSeeking);
	}

	/// *** CSourceSeeking *** ///
private:
	HRESULT STDMETHODCALLTYPE OnSeek(bool start)
	{
		HRESULT	hr	= S_OK;
		bool	z	= TRUE == ThreadExists();
		if(z) {
			if(S_OK != (hr = DeliverBeginFlush())) goto exit;
			if(S_OK != (hr = Stop())) goto exit;
			if(S_OK != (hr = DeliverEndFlush())) goto exit;
		}
		/// timestamps must restart from 0 after a seek, so keep track of first frame in m_k
		if(start) m_k = m_i = m_rtStart / m_tpf;
		if(z) {
			if(S_OK != (hr = Run())) goto exit;
		}
	exit:
		return hr;
	}
protected:
	HRESULT ChangeStart(void) { return OnSeek(true); }
	HRESULT ChangeStop(void) { return OnSeek(false); }
	HRESULT ChangeRate(void) { return OnSeek(false); }
};

class __declspec(uuid("{F8DDAE70-63C7-4b38-B7A9-61223765ABA6}")) CJpegPushSource : public CSource
{
friend HRESULT STDMETHODCALLTYPE CreateInstance<CJpegPushSource,IBaseFilter>(IBaseFilter** ppv);
private:
	CJpegPushSource(HRESULT* phr) : CSource(TEXT("CJpegPushSource"),NULL,__uuidof(CJpegPushSource),phr)
	{
		if(S_OK != *phr) return;

		HRESULT hr = S_OK;

		CJpegPushStream* pin = new CJpegPushStream(&hr,this);
		if(pin == NULL) { hr = E_OUTOFMEMORY; goto exit; }
		if(S_OK != hr) goto exit;

		pin = NULL;

	exit:
		if(pin) delete pin;
		*phr = hr;
	}
};


/// this demonstrates how to use DirectShow in a GUI-less console application
int main(int argc, LPCTSTR argv[])
{
	HRESULT						hr				= S_OK;
	bool						com				= false;
	IGraphBuilder*				pGrfBuilder		= NULL;
	IObjectWithSite*			pGrfSite		= NULL;
	IMediaControl*				pGrfControl		= NULL;
	IMediaSeeking*				pGrfSeeking		= NULL;
	IAMGraphBuilderCallback*	pGrfCallback	= NULL;
	IBaseFilter*				pSrcFilter		= NULL;
	IEnumPins*					pPins			= NULL;
	IPin*						pPin			= NULL;
	long						lEvent;

	if(FAILED(hr = CoInitialize(NULL))) goto exit; else com = true;

	if(S_OK != (hr = CoCreateInstance(CLSID_FilterGraph,NULL,CLSCTX_INPROC_SERVER,__uuidof(IGraphBuilder),(void**)&pGrfBuilder))) goto exit;

	/// install our graph builder callback to reject unwanted filters
	if(S_OK != (hr = CreateInstance<CSomeDevelopersMustDieGraphBuilderCallback,IAMGraphBuilderCallback>(&pGrfCallback))) goto exit;
	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IObjectWithSite),(void**)&pGrfSite))) goto exit;
	if(S_OK != (hr = pGrfSite->SetSite(pGrfCallback))) goto exit; else { pGrfSite->Release(); pGrfSite = NULL; }

	/// create our internal filter
	if(S_OK != (hr = CreateInstance<CJpegPushSource,IBaseFilter>(&pSrcFilter))) goto exit;
	if(S_OK != (hr = pGrfBuilder->AddFilter(pSrcFilter,NULL))) goto exit;

	/// use IntelligentConnect to render the filter's output pins, which in our case builds a complete rendering graph
	if(S_OK != (hr = pSrcFilter->EnumPins(&pPins))) goto exit;
	while(S_OK == (hr = pPins->Next(1,&pPin,NULL))) {
		PIN_DIRECTION dir; IPin* pin;
		if(S_OK == (hr = pPin->ConnectedTo(&pin))) { pin->Release(); goto next; } else if(VFW_E_NOT_CONNECTED != hr) goto exit;
		if(S_OK != (hr = pPin->QueryDirection(&dir))) goto exit; else if(dir != PINDIR_OUTPUT) goto next;
		if(S_OK != (hr = pGrfBuilder->Render(pPin))) goto exit;
	next:
		pPin->Release(); pPin = NULL;
	} if(S_FALSE != hr) goto exit; else { hr = S_OK; pPins->Release(); pPins = NULL; }

	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IMediaControl	),(void**)&pGrfControl	))) goto exit;
	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IMediaSeeking	),(void**)&pGrfSeeking	))) goto exit;

	if(S_OK != (hr = pGrfControl->Run()) && S_FALSE != hr) goto exit;

	REFERENCE_TIME rt;

	/// test seeking
	if(S_OK != (hr = WaitForCompletion(pGrfBuilder,1000,&lEvent,NULL)) && E_ABORT != hr) goto exit;
	if(S_OK != (hr = pGrfSeeking->SetPositions(((rt =    0 * REFTIME_UNITS_PER_MILLISEC),&rt),AM_SEEKING_AbsolutePositioning,NULL,AM_SEEKING_NoPositioning)) && S_FALSE != hr) goto exit;
	if(S_OK != (hr = WaitForCompletion(pGrfBuilder,1000,&lEvent,NULL)) && E_ABORT != hr) goto exit;
	if(S_OK != (hr = pGrfSeeking->SetPositions(((rt = 3000 * REFTIME_UNITS_PER_MILLISEC),&rt),AM_SEEKING_AbsolutePositioning,NULL,AM_SEEKING_NoPositioning)) && S_FALSE != hr) goto exit;
	if(S_OK != (hr = WaitForCompletion(pGrfBuilder,INFINITE,&lEvent,NULL)) && E_ABORT != hr) goto exit;

	hr = S_OK;
exit:
	if(pPin			) pPin			->Release();
	if(pPins		) pPins			->Release();
	if(pSrcFilter	) pSrcFilter	->Release();
	if(pGrfControl	) pGrfControl	->Release();
	if(pGrfSeeking	) pGrfSeeking	->Release();
	if(pGrfSite		) pGrfSite		->Release();
	if(pGrfBuilder	) pGrfBuilder	->Release();
	if(pGrfCallback	) pGrfCallback	->Release();
	if(com			) CoUninitialize();
	return hr;
}
