/// making valid JFIF files is not necessary to make JPEGs that can be read by smart JPEG viewers
#define MAKE_VALID_JFIF

#include "utils.h"
#include <qedit.h>
#define MJPGDHTSEG_STORAGE /// necessary to define MJPGDHTSeg
#define MJPGHDTSEG_STORAGE /// necessary because of a mispelling in mmreg.h
#include <mmreg.h>
#include <stdio.h>


/// just some useful definitions for JPEG and JFIF markers
#ifdef MAKE_VALID_JFIF

/// JPEG is big-endian, while the CPUs Windows[Mobile] currently supports are little-endian
#define SWAPWORDBYTES(w) ((WORD)((WORD)(BYTE)((w) >> 8) | ((w) << 8)))

#define JPEG_DHT	0xFFC4
#define JPEG_SOI	0xFFD8
#define JPEG_EOI	0xFFD9
#define JPEG_SOS	0xFFDA
#define JPEG_DQT	0xFFDB
#define JPEG_APP0	0xFFE0

#define JFIF_UNITS_NONE 0

#pragma pack(1)
typedef struct JFIFAPPHDR {
	WORD	soi, app0;
	WORD	len;
	BYTE	jfif[5];
	BYTE	maj_ver, min_ver;
	BYTE	units;
	WORD	ar_x, ar_y;
	BYTE	thumb_x, thumb_y;
} JFIFAPPHDR, *LPJFIFAPPHDR;
#pragma pack()

JFIFAPPHDR JFIFAppHdr = {
	SWAPWORDBYTES(JPEG_SOI), SWAPWORDBYTES(JPEG_APP0),
	SWAPWORDBYTES((WORD)((LPBYTE)&(&JFIFAppHdr)[1] - (LPBYTE)&JFIFAppHdr.len)),
	"JFIF",
	1,1,
	JFIF_UNITS_NONE,
	SWAPWORDBYTES(1),SWAPWORDBYTES(1),
	0,0
};

#endif /// MAKE_VALID_JFIF


class CSampleGrabberCallback : public CUnknown, public ISampleGrabberCB
{
friend HRESULT STDMETHODCALLTYPE CreateInstance<CSampleGrabberCallback,ISampleGrabberCB>(ISampleGrabberCB** ppv);
private:
	CSampleGrabberCallback(HRESULT* phr) : CUnknown(TEXT("CSampleGrabberCallback"),NULL,phr) {}

	/// *** INonDelegatingUnknown *** ///
protected:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(riid == __uuidof(ISampleGrabberCB)) return GetInterface((ISampleGrabberCB*)this,ppv);
		return CUnknown::NonDelegatingQueryInterface(riid,ppv);
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

	/// *** ISampleGrabberCB *** ///
public:
	HRESULT STDMETHODCALLTYPE SampleCB(double secs, IMediaSample *pms)
	{
		HRESULT			hr		= S_OK;
		DWORD			rc		= ERROR_SUCCESS;
		HANDLE			hf		= INVALID_HANDLE_VALUE;
		LPBYTE			pb;
		DWORD			cb, rb, wb;
		REFERENCE_TIME	rt1, rt2;
		TCHAR			fn[MAX_PATH];

		if(S_OK != (hr = pms->GetPointer(&pb))) goto exit; else cb = (DWORD)pms->GetActualDataLength();
		/// if the sample has no timestamp, ignore it since we use the timestamp as file name
		if(S_OK != (hr = pms->GetTime(&rt1,&rt2)) && VFW_S_NO_STOP_TIME != hr) goto exit;

		/// secs is adjusted by the seeking offset, while the timestamp is zero-based relative to the current segment
		rt1 = (REFERENCE_TIME)(secs * REFTIME_UNITS_PER_SEC);

		(void)
#ifdef UNICODE
			swprintf_s
#else
			sprintf_s
#endif
		/// change this to change the naming scheme of the JPEG files
		(fn,TEXT("target\\%06I64d.%07I64d.jpg"),rt1/REFTIME_UNITS_PER_SEC,rt1%REFTIME_UNITS_PER_SEC);

#ifdef MAKE_VALID_JFIF
		/// skipping SOI and APP0 markers
		for(LPBYTE z = (LPBYTE)-1i64, p = pb, q = &p[cb - 1]; z && p < q; p++) {
			if(*p == (BYTE)0xFF) {
				switch(p[1])
				{
					case (BYTE)0xD8: /// SOI, JFIFAppHdr has its own SOI 
					case (BYTE)0xE0: /// APP0, should be 'AVI1'
						break;
					default:
						z = NULL; cb = (DWORD)(&pb[cb] - p); pb = p;
				}
			}
		}
#endif /// MAKE_VALID_JFIF

		/// looking for SOS marker
		LPBYTE pk = NULL;
		for(DWORD k = 0, n = cb - 2; k < n; k++) if(pb[k] == (BYTE)0xFF && pb[k+1] == (BYTE)0xDA) { pk = &pb[k]; break; }
		if(pk == NULL) pk = pb;

		/// change this block to store the JPEG data in some other way
		{
			if(INVALID_HANDLE_VALUE == (hf = CreateFile(fn,FILE_WRITE_DATA,FILE_SHARE_READ,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL))) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; }

#ifdef MAKE_VALID_JFIF
			/// writing the JFIF header, including the leading SOI
			if(!WriteFile(hf,&JFIFAppHdr,rb = (DWORD)sizeof(JFIFAppHdr),&wb,NULL)) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; } else if(wb != rb) { hr = E_UNEXPECTED; goto exit; }
#endif /// MAKE_VALID_JFIF
			/// writing everything up to the SOS
			if(pk > pb) if(!WriteFile(hf,pb,rb = (DWORD)(pk-pb),&wb,NULL)) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; } else if(wb != rb) { hr = E_UNEXPECTED; goto exit; }
			/// writing the MJPGDHTSeg table, which is implicit in MJPEG but not JPEG, and which starts with a DHT marker
			if(!WriteFile(hf,MJPGDHTSeg,rb = (DWORD)sizeof(MJPGDHTSeg),&wb,NULL)) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; } else if(wb != rb) { hr = E_UNEXPECTED; goto exit; }
			/// writing the rest of the JPEG data starting at the SOS
			if(!WriteFile(hf,pk,rb = (DWORD)(&pb[cb]-pk),&wb,NULL)) { hr = HRESULT_FROM_WIN32(GetLastError()); goto exit; } else if(wb != rb) { hr = E_UNEXPECTED; goto exit; }
		}

	exit:
		if(hf != INVALID_HANDLE_VALUE) (void)CloseHandle(hf);
		return hr;
	}
	HRESULT STDMETHODCALLTYPE BufferCB(double secs, LPBYTE pb, long cb)
	{
		return E_NOTIMPL;
	}
};


/// this demonstrates how to use DirectShow in a GUI-less console application
int main(int argc, LPCTSTR argv[])
{
	HRESULT						hr				= S_OK;
	bool						com				= false;
	IGraphBuilder*				pGrfBuilder		= NULL;
	IObjectWithSite*			pGrfSite		= NULL;
	IMediaFilter*				pGrfFilter		= NULL;
	IMediaControl*				pGrfControl		= NULL;
	IAMGraphBuilderCallback*	pGrfCallback	= NULL;
	IBaseFilter*				pSrcFilter		= NULL;
	IPin*						pSrcPin			= NULL;
	IMediaSeeking*				pSrcSeeking		= NULL;
	IBaseFilter*				pEndFilter		= NULL;
	ISampleGrabber*				pEndGrabber		= NULL;
	ISampleGrabberCB*			pEndCallback	= NULL;
	IPin*						pEndPin			= NULL;
	AM_MEDIA_TYPE				mt				= { MEDIATYPE_Video, MEDIASUBTYPE_MJPG, FALSE, FALSE, 0, FORMAT_VideoInfo, NULL, 0, NULL };
	long						lEvent;

	if(FAILED(hr = CoInitialize(NULL))) goto exit; else com = true;

	if(S_OK != (hr = CoCreateInstance(CLSID_FilterGraph,NULL,CLSCTX_INPROC_SERVER,__uuidof(IGraphBuilder),(void**)&pGrfBuilder))) goto exit;

	/// install our graph builder callback to reject unwanted filters
	if(S_OK != (hr = CreateInstance<CSomeDevelopersMustDieGraphBuilderCallback,IAMGraphBuilderCallback>(&pGrfCallback))) goto exit;
	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IObjectWithSite),(void**)&pGrfSite))) goto exit;
	if(S_OK != (hr = pGrfSite->SetSite(pGrfCallback))) goto exit; else { pGrfSite->Release(); pGrfSite = NULL; }

	/// add the sample grabber configured to accept MJPEG video and to work in sample mode
	if(S_OK != (hr = CoCreateInstance(CLSID_SampleGrabber,NULL,CLSCTX_INPROC_SERVER,__uuidof(IBaseFilter),(void**)&pEndFilter))) goto exit;
	if(S_OK != (hr = pGrfBuilder->AddFilter(pEndFilter,NULL))) goto exit;
	if(S_OK != (hr = pEndFilter->QueryInterface(__uuidof(ISampleGrabber),(void**)&pEndGrabber))) goto exit;
	if(S_OK != (hr = pEndGrabber->SetMediaType(&mt))) goto exit;
	if(S_OK != (hr = CreateInstance<CSampleGrabberCallback,ISampleGrabberCB>(&pEndCallback))) goto exit;
	if(S_OK != (hr = pEndGrabber->SetCallback(pEndCallback,0))) goto exit;
	if(S_OK != (hr = FindFirstPin(pEndFilter,false,PINDIR_INPUT,GUID_NULL,&pEndPin,MEDIATYPE_NULL))) goto exit;

	/// let IntelligentConnect find an appropriate source filter
	if(S_OK != (hr = pGrfBuilder->AddSourceFilter(L"source.avi",NULL,&pSrcFilter))) goto exit;
	if(S_OK != (hr = FindFirstPin(pSrcFilter,false,PINDIR_OUTPUT,GUID_NULL,&pSrcPin,MEDIATYPE_Stream,MEDIATYPE_Video,MEDIATYPE_NULL))) goto exit;

	/// let IntelligentConnect build a filter chain between the source and the grabber
	if(S_OK != (hr = pGrfBuilder->Connect(pSrcPin,pEndPin))) goto exit;

	/// let's see if our upstream filter (most likely a parser) can seek
	pSrcPin->Release(); pSrcPin = NULL;
	if(S_OK != (hr = pEndPin->ConnectedTo(&pSrcPin))) goto exit;
	if(S_OK != (hr = pSrcPin->QueryInterface(__uuidof(IMediaSeeking),(void**)&pSrcSeeking))) goto exit;
	pSrcPin->Release(); pSrcPin = NULL;

	/// let IntelligentConnect build the rest of rendering graph
	/// in real-life, it may be smarter to use a NullRenderer
	pEndPin->Release(); pEndPin = NULL;
	if(S_OK != (hr = FindFirstPin(pEndFilter,false,PINDIR_OUTPUT,GUID_NULL,&pEndPin,MEDIATYPE_NULL))) goto exit;
	if(S_OK != (hr = pGrfBuilder->Render(pEndPin))) goto exit;
	pEndPin->Release(); pEndPin = NULL;

	/// removing the clock to let the graph run as fast as possible (necessary since we have a real renderer)
	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IMediaFilter),(void**)&pGrfFilter))) goto exit;
	if(S_OK != (hr = pGrfFilter->SetSyncSource(NULL))) goto exit;

	if(S_OK != (hr = pGrfBuilder->QueryInterface(__uuidof(IMediaControl	),(void**)&pGrfControl	))) goto exit;

	/// if our parser can seek, let's extract only frames in between 1 and 2 secs
	if(pSrcSeeking) {
		REFERENCE_TIME rt1 = 1000 * REFTIME_UNITS_PER_MILLISEC, rt2 = 2000 * REFTIME_UNITS_PER_MILLISEC;
		if(S_OK != (hr = pSrcSeeking->SetPositions(&rt1,AM_SEEKING_AbsolutePositioning,&rt2,AM_SEEKING_AbsolutePositioning)) && S_FALSE != hr) goto exit;
	}

	if(S_OK != (hr = pGrfControl->Run()) && S_FALSE != hr) goto exit;

	if(S_OK != (hr = WaitForCompletion(pGrfBuilder,INFINITE,&lEvent,NULL)) && E_ABORT != hr) goto exit;

	hr = S_OK;
exit:
	if(pSrcSeeking	) pSrcSeeking	->Release();
	if(pSrcPin		) pSrcPin		->Release();
	if(pSrcFilter	) pSrcFilter	->Release();
	if(pEndPin		) pEndPin		->Release();
	if(pEndFilter	) pEndFilter	->Release();
	if(pEndGrabber	) pEndGrabber	->Release();
	if(pEndCallback	) pEndCallback	->Release();
	if(pGrfControl	) pGrfControl	->Release();
	if(pGrfFilter	) pGrfFilter	->Release();
	if(pGrfSite		) pGrfSite		->Release();
	if(pGrfBuilder	) pGrfBuilder	->Release();
	if(pGrfCallback	) pGrfCallback	->Release();
	if(com			) CoUninitialize();
	return hr;
}
